var sys = require("sys");
var mysql = require("mysql");

var MysqlClient = function(configuration, internalClient) {
  this.configuration = configuration;
  this.internalClient = internalClient;
  this.waitForCompletion = false;
  this.executingRequestNumber = 0;
  this.internalClient.on("error", function(err) {
      console.log(sys.inspect(err));
  });
}

MysqlClient.prototype = {
    setErrorHandler: function(errorHandler) {
        this.errorHandler = errorHandler;
        if (this.internalClient!=null) {
            this.internalClient.on("error", errorHandler);
        }
    },

    beginTransaction: function() {
        this.internalClient.query("BEGIN");
    },

    commit: function() {
        this.internalClient.query("COMMIT");
    },

    rollback: function() {
        this.internalClient.query("ROLLBACK");
    },

    createQuery: function(queryName, query) {
        var queryString = this._convertQueryString(query);
        return new MysqlQuery(this, queryName, queryString);
    },

    _convertDataAccessError: function(err) {
        //TODO: add error conversion
        return err;
    },

    _handleErrorIfNecessary: function(err) {
        if (this.errorHandler!=null) {
            this.errorHandler(this._convertDataAccessError(err));
        }
    },

    enableWaitForCompletion: function(onCompleteCallback) {
        this.waitForCompletion = true;
        this.onCompleteCallback = onCompleteCallback;
    },

    _triggerRequestStarted: function() {
        this.executingRequestNumber++;
    },

    _triggerRequestCompleted: function() {
        this.executingRequestNumber--;
        if (this.waitForCompletion && this.executingRequestNumber==0) {
            if (this.onCompleteCallback!=null) {
                this.onCompleteCallback();
            }
            if (!this.pooling) {
                this.internalClient.end();
            }
        }
    },

    getTableMetadata: function(resultCallback) {
        var self = this;
        this.executeQuery("show tables", function(result) {
            var tableNames = [];
            for (var i=0; i<result.length; i++) {
                var elt = result[i];
                var tableData = {}
                tableData.name = elt["Tables_in_"+self.configuration.database];
                tableNames.push(tableData);
            }

            resultCallback(tableNames);
        });
    },

    getColumnMetadata: function(tableName, resultCallback) {
        this.executeQuery("show columns from "+tableName,
                      function(result) {
            var columns = [];
            for (var i=0; i<result.length; i++) {
                var elt = result[i];
                var fieldData = {};
                fieldData.name = elt["Field"];
                fieldData.type = elt["Type"];
                fieldData.null = (elt["Null"]=="YES");
                fieldData.default = elt["Default"];
                columns.push(fieldData);
            }
            resultCallback(columns);
        });
    },

    _convertQueryString: function(q) {
        //TODO: improve the parameter conversion
        /*for (var i=0; i<q.length; i++) {
            var c = q.charAt(i);
            
        }*/
        for (var i=0;i<10;i++) {
            q = q.replace("$"+i, "?");
        }
        return q;
    },

    executeQuery: function() {
        var queryString = this._convertQueryString(arguments[0]);
        var parameters = arguments.length==3 ? arguments[1] : null;
        var resultCallback = arguments.length==3 ? arguments[2] : arguments[1];
        var self = this;
        if (parameters!=null) {
            self._triggerRequestStarted();
            self.internalClient.query(queryString, parameters, function(err, results, fields) {
                if (err!=null) {
                    self._handleErrorIfNecessary(err);
                } else {
                    resultCallback(results);
                }
                self._triggerRequestCompleted();
            });
        } else {
            self._triggerRequestStarted();
            var query = self.internalClient.query(queryString, function(err, results, fields) {
                if (err!=null) {
                    self._handleErrorIfNecessary(err);
                } else {
                    resultCallback(results);
                }
                self._triggerRequestCompleted();
            });
        }
    },

    executeUpdate: function(query, parameters) {
        var queryString = this._convertQueryString(arguments[0]);
        var parameters = arguments.length==3 ? arguments[1] : null;
        var resultCallback = arguments.length==3 ? arguments[2] : arguments[1];
        var self = this;
        if (parameters!=null) {
            self._triggerRequestStarted();
            var query = this.internalClient.query(query, parameters);
            query.on("error", function(err) {
                self._handleErrorIfNecessary(err);
            });
            query.on("end", function() {
                self._triggerRequestCompleted();
            });
        } else {
            var query = this.internalClient.query(query);
            query.on("error", function(err) {
                self._handleErrorIfNecessary(err);
            });
            query.on("end", function() {
                self._triggerRequestCompleted();
            });
        }
    },

    end: function() {
        console.log("end");
        this.internalClient.end();
    },

    finalizePool: function() {
    }
};

var MysqlQuery = function(client, queryName, query) {
    this.client = client;
    this.queryName = queryName;
    this.query = query;
}

MysqlQuery.prototype = {
    execute: function() {
        var parameters = null;
        var resultCallback = null;
        if (arguments.length==2) {
            parameters = arguments[0];
            resultCallback = arguments[1];
        } else if (arguments.length==1) {
            if (typeof arguments[0] == "function") {
                resultCallback = arguments[0];
            } else if (arguments[0] instanceof Array) {
                parameters = arguments[0];
            }
        }

        var self = this;
        if (resultCallback!=null) {
            self.client._triggerRequestStarted();
            var queryCallback = function(err, results, fields) {
                self.client._handleErrorIfNecessary(err);
                if (!err) {
                    resultCallback(results);
                }
                self.client._triggerRequestCompleted();
            };
            if (parameters!=null) {
                this.internalQuery = this.client.internalClient.query(this.query, parameters, queryCallback);
            } else {
                this.internalQuery = this.client.internalClient.query(this.query, queryCallback);
            }
        } else {
            self.client._triggerRequestStarted();
            if (parameters!=null) {
                this.internalQuery = this.client.internalClient.query(this.query, parameters);
            } else {
                this.internalQuery = this.client.internalClient.query(this.query);
            }
            this.internalQuery.on("end", function() {
                self.client._triggerRequestCompleted();
            });
        }
    },
    listen: function(eventName, callback) {
        //suports event: result, row, end
        this.internalQuery.on(eventName, callback);
    }
};

var MysqlDatabaseTemplate = function(configuration, callbacks) {
    this.configuration = configuration;
    this.callbacks = callbacks;
}

MysqlDatabaseTemplate.prototype = {
    _getOnCompleteCallback: function() {
        return (this.callbacks!=null && this.callbacks.onComplete
                 ? this.callbacks.onComplete : null);
    },

    _getErrorHandlerCallback: function() {
        return (this.callbacks!=null && this.callbacks.errorHandler
                 ? this.callbacks.errorHandler : null);
    },

    doOnData: function(callback) {
        if (this.configuration.pooling) {
            //TODO: implement pool approach
            throw new Error("Pooling approach isn't supported the MySQL abstraction implementation.");
        } else {
            var config = {};
            config.user = this.configuration.user;
            config.password = this.configuration.password;
            if (this.configuration.port) {
                config.port = this.configuration.port;
            }
            if (this.configuration.host) {
                config.host = this.configuration.host;
            }
            var internalClient = mysql.createClient(config);
            internalClient.query("USE "+this.configuration.database);
            var client = new MysqlClient(this.configuration, internalClient);
            if (this._getErrorHandlerCallback()!=null) {
                client.setErrorHandler(this._getErrorHandlerCallback());
            }
            callback(client);
            client.enableWaitForCompletion(this._getOnCompleteCallback());
        }
    }
};

module.exports = {
    DatabaseTemplate: MysqlDatabaseTemplate
};
