var sys = require("sys");
var pgjs = require("pg");
var pgnative = require("pg").native;
var genericdriver = require("./driver");

var PostgresqlClient = function(configuration, internalClient) {
  this.configuration = configuration;
  this.internalClient = internalClient;
  this.waitForCompletion = false;
  this.executingRequestNumber = 0;
  this.internalClient.on("error", function(err) {
      console.log(sys.inspect(err));
  });
}

PostgresqlClient.prototype = {
    setErrorHandler: function(errorHandler) {
        this.errorHandler = errorHandler;
        if (this.internalClient!=null) {
            this.internalClient.on("error", errorHandler);
        }
    },

    beginTransaction: function() {
        this.internalClient.query("BEGIN");
    },

    commit: function() {
        this.internalClient.query("COMMIT");
    },

    rollback: function() {
        this.internalClient.query("ROLLBACK");
    },

    createQuery: function(queryName, query) {
        return new PostgresqlQuery(this, queryName, query);
    },

    _convertDataAccessError: function(err) {
        //TODO: add error conversion
        return err;
    },

    _handleErrorIfNecessary: function(err) {
        if (this.errorHandler!=null) {
            this.errorHandler(this._convertDataAccessError(err));
        } else {
            throw err;
        }
    },

    enableWaitForCompletion: function(onCompleteCallback) {
        this.waitForCompletion = true;
        this.onCompleteCallback = onCompleteCallback;
    },

    _triggerRequestStarted: function() {
        this.executingRequestNumber++;
    },

    _triggerRequestCompleted: function() {
        this.executingRequestNumber--;
        if (this.waitForCompletion && this.executingRequestNumber==0) {
            this.internalClient.end();

            if (this.onCompleteCallback!=null) {
                this.onCompleteCallback(this);
            }
            this.waitForCompletion = false;
        }
    },

    getTableMetadata: function(resultCallback) {
        this.executeQuery("SELECT tablename FROM pg_tables" +
                      " WHERE tablename NOT LIKE 'pg\\_%'" +
                      " AND tablename NOT LIKE 'sql\\_%'", function(result) {
            var tables = [];
            for (var i=0; i<result.length; i++) {
                var elt = result[i];
                tables.push(elt["tablename"]);
            }
            resultCallback(tables);
        });
    },

    getColumnMetadata: function(tableName, resultCallback) {
        this.executeQuery("SELECT column_name, data_type, is_nullable, column_default," +
                      "character_maximum_length FROM information_schema.COLUMNS" +
                      " WHERE table_name = $1",
                      [ tableName ], function(result) {
            var columns = [];
            for (var i=0; i<result.length; i++) {
                var elt = result[i];
                var fieldData = {};
                fieldData.name = elt["column_name"];
                fieldData.type = elt["data_type"];
                fieldData.maxLength = elt["character_maximum_length"];
                fieldData.null = elt["is_nullable"];
                fieldData.default = elt["column_default"];
                columns.push(fieldData);
            }
            resultCallback(columns);
        });
    },

    executeQuery: function() {
        var queryString = arguments[0];
        var parameters = arguments.length==3 ? arguments[1] : null;
        var resultCallback = arguments.length==3 ? arguments[2] : arguments[1];
        var self = this;
        if (parameters!=null) {
            self._triggerRequestStarted();
            self.internalClient.query({
                text: queryString,
                values: parameters
            }, function(err, result) {
                if (err!=null) {
                    self._handleErrorIfNecessary(err);
                } else {
                    resultCallback(result.rows);
                }
                self._triggerRequestCompleted();
            });
        } else {
            self._triggerRequestStarted();
            var query = self.internalClient.query(queryString, function(err, result) {
                if (err!=null) {
                    self._handleErrorIfNecessary(err);
                } else {
                    resultCallback(result.rows);
                }
                self._triggerRequestCompleted();
            });
        }
    },

    executeUpdate: function(query, parameters) {
        var self = this;
        if (parameters!=null) {
            self._triggerRequestStarted();
            var query = this.internalClient.query(query, parameters);
            query.on("error", function(err) {
                self._handleErrorIfNecessary(err);
            });
            query.on("end", function() {
                self._triggerRequestCompleted();
            });
        } else {
            self._triggerRequestStarted();
            var query = this.internalClient.query(query);
            query.on("error", function(err) {
                self._handleErrorIfNecessary(err);
            });
            query.on("end", function() {
                self._triggerRequestCompleted();
            });
        }
    },

    end: function() {
        this.internalClient.end();
    },

    finalizePool: function() {
        if (this.configuration.pooling) {
            var pg = this.configuration.native ? pgnative : pgjs;
            pg.end();
        }
    }
};

var PostgresqlQuery = function(client, queryName, query) {
    this.client = client;
    this.queryName = queryName;
    this.query = query;
}

PostgresqlQuery.prototype = {
    execute: function() {
        var parameters = null;
        var resultCallback = null;
        if (arguments.length==2) {
            parameters = arguments[0];
            resultCallback = arguments[1];
        } else if (arguments.length==1) {
            if (typeof arguments[0] == "function") {
                resultCallback = arguments[0];
            } else if (arguments[0] instanceof Array) {
                parameters = arguments[0];
            }
        }

        var self = this;
        var queryConfig = {};
        queryConfig["name"] = this.queryName;
        queryConfig["text"] = this.query;
        if (parameters!=null) {
            queryConfig["values"] = parameters;
        }
        if (resultCallback!=null) {
            self.client._triggerRequestStarted();
            this.client.internalQuery = this.client.internalClient.query(queryConfig, function(err, result) {
                self.client._handleErrorIfNecessary(err);
                if (!err) {
                    resultCallback(result.rows);
                }
                self.client._triggerRequestCompleted();
            });
        } else {
            self.client._triggerRequestStarted();
            this.internalQuery = this.client.internalClient.query(queryConfig);
            this.internalQuery.on("err", function() {
                self.client._handleErrorIfNecessary(err);
            });
            this.internalQuery.on("end", function() {
                self.client._triggerRequestCompleted();
            });
        }
    },
    listen: function(eventName, callback) {
        //suports event: result, row, end
        if (eventName=="result" || eventName=="row" || eventName=="end") {
            this.internalQuery.on(eventName, callback);
        }
    }
};

var PostgresqlDatabaseTemplate = function(configuration, callbacks) {
    this.configuration = configuration;
    this.callbacks = callbacks;
}

PostgresqlDatabaseTemplate.prototype = {
    _getOnCompleteCallback: function() {
        return (this.callbacks!=null && this.callbacks.onComplete
                 ? this.callbacks.onComplete : null);
    },

    _getErrorHandlerCallback: function() {
        return (this.callbacks!=null && this.callbacks.errorHandler
                 ? this.callbacks.errorHandler : null);
    },

    _handleErrorIfNecessary: function(err) {
        if (err!=null) {
            var errorHandler = this._getErrorHandlerCallback();
            if (errorHandler!=null) {
                errorHandler(err);
            } else {
                throw new Error(err);
            }
        }
    },

    doOnData: function(callback) {
        var self = this;
        var pg = this.configuration.native ? pgnative : pgjs;
        if (this.configuration.pooling) {
            pg.connect({
                user: this.configuration.user,
                database: this.configuration.database,
                password: this.configuration.password,
                port: this.configuration.port,
                host: this.configuration.host
            }, function(err, internalClient) {
                self._handleErrorIfNecessary(err);

                if (err==null) {
                    var client = new PostgresqlClient(self.configuration, internalClient);
                    if (self._getErrorHandlerCallback()!=null) {
                        client.setErrorHandler(self._getErrorHandlerCallback());
                    }
                    callback(client);
                    client.enableWaitForCompletion(self._getOnCompleteCallback());
                }
            });
        } else {
            var internalClient = new pg.Client({
                user: this.configuration.user,
                database: this.configuration.database,
                password: this.configuration.password,
                port: this.configuration.port,
                host: this.configuration.host
            });
            var client = new PostgresqlClient(this.configuration, internalClient);
            if (this._getErrorHandlerCallback()!=null) {
                client.setErrorHandler(this._getErrorHandlerCallback());
            }
            internalClient.connect();
            callback(client);
            client.enableWaitForCompletion(this._getOnCompleteCallback());
        }
    }
};

module.exports = {
    DatabaseTemplate: PostgresqlDatabaseTemplate
};
